"""Convert a segy file to tilesegy format"""

import shutil
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser, RawTextHelpFormatter
from pathlib import Path
from typing import List, Optional

import segyio
import tiledb

from .convert import SegyFileConverter


class HelpFormatter(RawTextHelpFormatter, ArgumentDefaultsHelpFormatter):
    pass


def get_parser() -> ArgumentParser:
    parser = ArgumentParser(formatter_class=HelpFormatter, description=__doc__)
    parser.add_argument("input", type=Path, help="Input segy file path")
    parser.add_argument(
        "output", type=Path, nargs="?", help="Output tilesegy directory path"
    )
    parser.add_argument(
        "-o",
        "--overwrite",
        action="store_true",
        help="Overwrite the output directory if it already exists",
    )
    parser.add_argument(
        "-g",
        "--geometry",
        choices=["auto", "structured", "unstructured"],
        default="auto",
        help="""Geometry of the converted tilesegy:
- auto: same as the input segy.
- structured: same as `auto` but abort if a geometry cannot be inferred.
- unstructured: opt out on building geometry information.
""",
    )

    segyio_args = parser.add_argument_group("segyio options")
    segyio_args.add_argument(
        "--su",
        action="store_true",
        help="Open a seismic unix file instead of SEG-Y",
    )
    segyio_args.add_argument(
        "--iline",
        type=int,
        default=189,
        help="Inline number field in the trace headers",
    )
    segyio_args.add_argument(
        "--xline",
        type=int,
        default=193,
        help="Crossline number field in the trace headers",
    )
    segyio_args.add_argument(
        "--endian",
        choices=["big", "msb", "little", "lsb"],
        default="big",
        help="File endianness, big/msb (default) or little/lsb",
    )

    tiledb_args = parser.add_argument_group("tiledb options")
    tiledb_args.add_argument(
        "-s",
        "--tile-size",
        type=int,
        default=4_000_000,
        help="Tile size in bytes.\n"
        "Larger tile size improves disk access time at the cost of higher memory",
    )
    tiledb_args.add_argument(
        "--consolidation-buffersize",
        type=int,
        default=5_000_000,
        help="The size in bytes of the attribute buffers used during consolidation",
    )

    return parser


def main(argv: Optional[List[str]] = None) -> None:
    parser = get_parser()
    args = parser.parse_args(argv)

    if not args.output:
        args.output = args.input.with_suffix(".tsgy")
    if args.output.exists():
        if args.overwrite:
            shutil.rmtree(args.output)
        else:
            parser.error(f"{args.output} already exists")

    segyio_kwargs = dict(
        filename=args.input,
        iline=args.iline,
        xline=args.xline,
        endian=args.endian,
    )
    if args.geometry == "unstructured":
        segyio_kwargs["ignore_geometry"] = True
    else:
        segyio_kwargs["strict"] = args.geometry != "auto"

    converter_kwargs = dict(
        tile_size=args.tile_size,
        config=tiledb.Config(
            {"sm.consolidation.buffer_size": args.consolidation_buffersize}
        ),
    )
    open_seismic = segyio.su.open if args.su else segyio.open
    with open_seismic(**segyio_kwargs) as f:
        converter = SegyFileConverter(f, **converter_kwargs)  # type: ignore
        converter.to_tiledb(args.output)


if __name__ == "__main__":
    main()



import copy
from abc import ABC, abstractmethod
from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional, Union, cast

import numpy as np
import segyio
import tiledb
from cached_property import cached_property

TypedTraceField = namedtuple("TypedTraceField", ["name", "enum", "dtype"])


def iter_typed_trace_fields() -> Iterator[TypedTraceField]:
    all_fields = segyio.TraceField.enums()
    include_names = set(map(str, segyio.field.Field._tr_keys))
    size2dtype = {2: np.dtype(np.int16), 4: np.dtype(np.int32)}
    for f, f2 in zip(all_fields, all_fields[1:]):
        name = str(f)
        if name in include_names:
            yield TypedTraceField(name, f, size2dtype[int(f2) - int(f)])


TRACE_FIELDS = tuple(iter_typed_trace_fields())
TRACE_FIELD_ENUMS = tuple(f.enum for f in TRACE_FIELDS)
TRACE_FIELD_NAMES = tuple(f.name for f in TRACE_FIELDS)
TRACE_FIELD_DTYPES = tuple(f.dtype for f in TRACE_FIELDS)
TRACE_FIELDS_SIZE = sum(dtype.itemsize for dtype in TRACE_FIELD_DTYPES)
TRACE_FIELD_FILTERS = (
    tiledb.BitWidthReductionFilter(),
    tiledb.ByteShuffleFilter(),
    tiledb.LZ4Filter(),
)


class ExtendedSegyFile(segyio.SegyFile):
    @cached_property
    def trace_size(self) -> int:
        return len(self._samples) * int(self._dtype.itemsize)

    @cached_property
    def fast_headerline(self) -> segyio.line.HeaderLine:
        return self._header.iline if self.is_inline else self._header.xline

    @cached_property
    def fast_lines(self) -> np.ndarray:
        return self._ilines if self.is_inline else self._xlines

    @cached_property
    def slow_lines(self) -> np.ndarray:
        return self._xlines if self.is_inline else self._ilines

    @cached_property
    def is_inline(self) -> bool:
        assert not self.unstructured
        return bool(self.sorting == segyio.TraceSortingFormat.INLINE_SORTING)


class SegyFileConverter(ABC):
    def __new__(cls, segy_file: segyio.SegyFile, **kwargs: Any) -> "SegyFileConverter":
        if cls is SegyFileConverter:
            if segy_file.unstructured:
                cls = UnstructuredSegyFileConverter
            else:
                cls = StructuredSegyFileConverter
        return cast(SegyFileConverter, super().__new__(cls))

    def __init__(
        self,
        segy_file: segyio.SegyFile,
        *,
        tile_size: int,
        config: Optional[tiledb.Config] = None,
    ):
        if not isinstance(segy_file, ExtendedSegyFile):
            segy_file = copy.copy(segy_file)
            segy_file.__class__ = ExtendedSegyFile
        self.segy_file = segy_file
        self.tile_size = tile_size
        self.config = config

    def to_tiledb(self, uri: Union[str, Path]) -> None:
        uri = Path(uri) if not isinstance(uri, Path) else uri

        if tiledb.object_type(str(uri)) != "group":
            tiledb.group_create(str(uri))

        headers_uri = uri / "headers"
        if tiledb.object_type(str(headers_uri)) != "array":
            dims = self._get_dims(TRACE_FIELDS_SIZE)
            header_schema = tiledb.ArraySchema(
                domain=tiledb.Domain(*dims),
                attrs=[
                    tiledb.Attr(f.name, f.dtype, filters=TRACE_FIELD_FILTERS)
                    for f in TRACE_FIELDS
                ],
            )
            with self._tiledb_array(str(headers_uri), header_schema) as tdb:
                self._fill_headers(tdb)

        data_uri = uri / "data"
        if tiledb.object_type(str(data_uri)) != "array":
            samples = len(self.segy_file.samples)
            sample_dtype = self.segy_file.dtype
            sample_size = sample_dtype.itemsize
            dims = list(self._get_dims(sample_size * samples))
            dims.append(
                tiledb.Dim(
                    name="samples",
                    domain=(0, samples - 1),
                    dtype=dims[0].dtype,
                    tile=np.clip(self.tile_size // sample_size, 1, samples),
                )
            )
            data_schema = tiledb.ArraySchema(
                domain=tiledb.Domain(*dims),
                attrs=[
                    tiledb.Attr("trace", sample_dtype, filters=(tiledb.LZ4Filter(),))
                ],
            )
            with self._tiledb_array(str(data_uri), data_schema) as tdb:
                self._fill_data(tdb)

    @contextmanager
    def _tiledb_array(
        self, uri: str, schema: tiledb.ArraySchema
    ) -> Iterator[tiledb.Array]:
        tiledb.DenseArray.create(uri, schema)
        with tiledb.DenseArray(uri, mode="w") as tdb:
            yield tdb
        tiledb.consolidate(uri, config=self.config)
        tiledb.vacuum(uri, config=self.config)

    @abstractmethod
    def _get_dims(self, trace_size: int) -> Iterable[tiledb.Dim]:
        """Get the tiledb schema dimensions"""

    @abstractmethod
    def _fill_headers(self, tdb: tiledb.Array) -> None:
        tdb.meta["__text__"] = b"".join(self.segy_file.text)
        for k, v in self.segy_file.bin.items():
            tdb.meta[str(k)] = v

    @abstractmethod
    def _fill_data(self, tdb: tiledb.Array) -> None:
        tdb.meta["sorting"] = (
            self.segy_file.sorting or segyio.TraceSortingFormat.UNKNOWN_SORTING
        )
        tdb.meta["samples"] = self.segy_file.samples.tolist()


class UnstructuredSegyFileConverter(SegyFileConverter):
    def _get_dims(self, trace_size: int) -> Iterable[tiledb.Dim]:
        traces = self.segy_file.tracecount
        return [
            tiledb.Dim(
                name="traces",
                domain=(0, traces - 1),
                dtype=np.uint64,
                tile=np.clip(self.tile_size // trace_size, 1, traces),
            ),
        ]

    def _fill_headers(self, tdb: tiledb.Array) -> None:
        super()._fill_headers(tdb)
        traces = self.segy_file.tracecount
        get_header = self.segy_file.header
        step = np.clip(self.tile_size // TRACE_FIELDS_SIZE, 1, traces)
        for sl in iter_slices(traces, step):
            headers = [
                np.zeros(sl.stop - sl.start, dtype) for dtype in TRACE_FIELD_DTYPES
            ]
            for i, field in enumerate(get_header[sl]):
                getfield, buf = field.getfield, field.buf
                for key, header in zip(TRACE_FIELD_ENUMS, headers):
                    v = getfield(buf, key)
                    if v:
                        header[i] = v
            tdb[sl] = dict(zip(TRACE_FIELD_NAMES, headers))

    def _fill_data(self, tdb: tiledb.Array) -> None:
        super()._fill_data(tdb)
        raw_trace = self.segy_file.trace.raw
        traces = self.segy_file.tracecount
        step = np.clip(self.tile_size // self.segy_file.trace_size, 1, traces)
        for sl in iter_slices(traces, step):
            tdb[sl] = raw_trace[sl]


class StructuredSegyFileConverter(SegyFileConverter):
    def _get_dims(self, trace_size: int) -> Iterable[tiledb.Dim]:
        dtype = np.uintc
        slow_lines = len(self.segy_file.slow_lines)
        if self.segy_file.is_inline:
            fast_dim, slow_dim = "ilines", "xlines"
        else:
            fast_dim, slow_dim = "xlines", "ilines"
        return [
            tiledb.Dim(
                name=fast_dim,
                domain=(0, len(self.segy_file.fast_lines) - 1),
                dtype=dtype,
                tile=self._fast_tile(trace_size),
            ),
            tiledb.Dim(
                name=slow_dim,
                domain=(0, slow_lines - 1),
                dtype=dtype,
                tile=slow_lines,
            ),
            tiledb.Dim(
                name="offsets",
                domain=(0, len(self.segy_file.offsets) - 1),
                dtype=dtype,
                tile=1,
            ),
        ]

    def _fill_headers(self, tdb: tiledb.Array) -> None:
        super()._fill_headers(tdb)
        step = self._fast_tile(TRACE_FIELDS_SIZE)
        fast_lines = self.segy_file.fast_lines
        slow_lines = self.segy_file.slow_lines
        fast_headerline = self.segy_file.fast_headerline
        for offset_idx, offset in enumerate(self.segy_file.offsets):
            for sl in iter_slices(len(fast_lines), step):
                slice_lines = fast_lines[sl]
                cubes = [
                    np.zeros((len(slice_lines), len(slow_lines)), dtype)
                    for dtype in TRACE_FIELD_DTYPES
                ]
                for i, line in enumerate(slice_lines):
                    for j, field in enumerate(fast_headerline[line, offset]):
                        getfield, buf = field.getfield, field.buf
                        for key, cube in zip(TRACE_FIELD_ENUMS, cubes):
                            v = getfield(buf, key)
                            if v:
                                cube[i, j] = v
                tdb[sl, :, offset_idx] = dict(zip(TRACE_FIELD_NAMES, cubes))

    def _fill_data(self, tdb: tiledb.Array) -> None:
        super()._fill_data(tdb)
        for key in "ilines", "xlines", "offsets":
            tdb.meta[key] = getattr(self.segy_file, key).tolist()

        step = self._fast_tile(self.segy_file.trace_size)
        fast_lines = self.segy_file.fast_lines
        get_line = self.segy_file.fast
        for offset_idx, offset in enumerate(self.segy_file.offsets):
            for sl in iter_slices(len(fast_lines), step):
                cube = np.stack([get_line[i, offset] for i in fast_lines[sl]])
                tdb[sl, :, offset_idx] = cube

    def _fast_tile(self, trace_size: int) -> int:
        num_fast, num_slow = map(
            len, (self.segy_file.fast_lines, self.segy_file.slow_lines)
        )
        return int(np.clip(self.tile_size // (num_slow * trace_size), 1, num_fast))


def iter_slices(size: int, step: int) -> Iterator[slice]:
    r = range(0, size, step)
    yield from map(slice, r, r[1:])
    yield slice(r[-1], size)



__all__ = ["open", "TileSegy", "StructuredTileSegy"]

from pathlib import Path
from types import TracebackType
from typing import TYPE_CHECKING, Optional, Tuple, Type, Union

import numpy as np
import tiledb
from segyio import TraceSortingFormat

if TYPE_CHECKING:  # pragma: nocover
    cached_property = property
else:
    from cached_property import cached_property

from .indexables import (
    Attributes,
    Depth,
    Field,
    Gather,
    Header,
    HeaderLine,
    Line,
    Trace,
)
from .utils import StructuredTraceIndexer, TraceIndexer


class TileSegy:
    _indexer_cls: Type[TraceIndexer] = TraceIndexer

    def __init__(self, uri: Path, headers: tiledb.Array, data: tiledb.Array):
        self._uri = uri
        self._headers = headers
        self._data = data

    @property
    def uri(self) -> Path:
        return self._uri

    @cached_property
    def sorting(self) -> Optional[TraceSortingFormat]:
        sorting = TraceSortingFormat(self._data.meta["sorting"])
        return sorting if sorting != TraceSortingFormat.UNKNOWN_SORTING else None

    @cached_property
    def bin(self) -> Field:
        bin_headers = dict(self._headers.meta.items())
        del bin_headers["__text__"]
        return bin_headers

    @cached_property
    def text(self) -> Tuple[bytes, ...]:
        text_headers = self._headers.meta["__text__"]
        assert len(text_headers) % 3200 == 0, len(text_headers)
        return tuple(
            text_headers[i : i + 3200] for i in range(0, len(text_headers), 3200)
        )

    @cached_property
    def samples(self) -> np.ndarray:
        return self._meta_to_numpy("samples")

    @cached_property
    def trace(self) -> Trace:
        return Trace(self._data, self._indexer_cls)

    @cached_property
    def header(self) -> Header:
        return Header(self._headers, self._indexer_cls)

    def attributes(self, name: str) -> Attributes:
        return Attributes(
            tiledb.DenseArray(self._headers.uri, attr=name), self._indexer_cls
        )

    @cached_property
    def depth_slice(self) -> Depth:
        return Depth(self._data)

    def close(self) -> None:
        self._headers.close()
        self._data.close()
        # remove all cached properties
        for attr in list(self.__dict__.keys()):
            if isinstance(getattr(self.__class__, attr, None), cached_property):
                delattr(self, attr)

    def __enter__(self) -> "TileSegy":
        return self

    def __exit__(
        self, type: Type[Exception], value: Exception, traceback: TracebackType
    ) -> None:
        self.close()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self._uri)!r})"

    def _meta_to_numpy(
        self, meta_key: str, dtype: Union[np.dtype, str, None] = None
    ) -> np.ndarray:
        values = self._data.meta[meta_key]
        if not isinstance(values, tuple):
            values = (values,)
        return np.array(values, dtype)


class StructuredTileSegy(TileSegy):
    _indexer_cls = StructuredTraceIndexer

    @cached_property
    def iline(self) -> Line:
        return Line("ilines", self.ilines, self.offsets, self._data)

    @cached_property
    def xline(self) -> Line:
        return Line("xlines", self.xlines, self.offsets, self._data)

    @cached_property
    def fast(self) -> Line:
        if self.sorting == TraceSortingFormat.INLINE_SORTING:
            return self.iline
        assert self.sorting == TraceSortingFormat.CROSSLINE_SORTING
        return self.xline

    @cached_property
    def slow(self) -> Line:
        if self.sorting == TraceSortingFormat.INLINE_SORTING:
            return self.xline
        assert self.sorting == TraceSortingFormat.CROSSLINE_SORTING
        return self.iline

    @cached_property
    def gather(self) -> Gather:
        return Gather(self.ilines, self.xlines, self.offsets, self._data)

    @cached_property
    def offsets(self) -> np.ndarray:
        return self._meta_to_numpy("offsets", dtype="intc")

    @cached_property
    def ilines(self) -> np.ndarray:
        return self._meta_to_numpy("ilines", dtype="intc")

    @cached_property
    def xlines(self) -> np.ndarray:
        return self._meta_to_numpy("xlines", dtype="intc")

    @cached_property
    def header(self) -> Header:
        header = super().header
        for attr, name in ("iline", "ilines"), ("xline", "xlines"):
            line = HeaderLine(name, getattr(self, name), self.offsets, self._headers)
            setattr(header, attr, line)
        return header


def open(uri: Union[str, Path]) -> TileSegy:
    uri = Path(uri) if not isinstance(uri, Path) else uri
    headers = tiledb.DenseArray(str(uri / "headers"))
    data = tiledb.DenseArray(str(uri / "data"), attr="trace")
    if data.schema.domain.has_dim("traces"):
        cls = TileSegy
    else:
        cls = StructuredTileSegy
    return cls(uri, headers, data)



from typing import Dict, List, Tuple, Type, Union, cast

import numpy as np
import tiledb

from .singledispatchmethod import singledispatchmethod  # type: ignore
from .utils import Index, Int, LabelIndexer, TraceIndexer, ensure_slice

Field = Dict[str, int]


class TraceIndexable:
    def __init__(self, tdb: tiledb.Array, indexer: TraceIndexer):
        self._tdb = tdb
        self._indexer = indexer

    def __len__(self) -> int:
        return len(self._indexer)


class Trace(TraceIndexable):
    def __init__(self, tdb: tiledb.Array, indexer_cls: Type[TraceIndexer]):
        super().__init__(tdb, indexer_cls(tdb.shape[:-1]))

    def __getitem__(self, i: Union[Index, Tuple[Index, Index]]) -> np.ndarray:
        if isinstance(i, tuple):
            # for single sample segyio returns an array of size 1 instead of scalar
            trace_index, samples = i[0], ensure_slice(i[1])
        else:
            trace_index, samples = i, slice(None)

        bounding_box, post_reshape_indices = self._indexer[trace_index]
        traces = self._tdb[(*bounding_box, samples)]
        if traces.ndim > 2:
            traces = traces.reshape(np.array(traces.shape[:-1]).prod(), -1)
        elif traces.size == 1:
            # convert to scalar (https://github.com/equinor/segyio/issues/475)
            post_reshape_indices = 0
        return traces[post_reshape_indices]


class Header(TraceIndexable):
    def __init__(self, tdb: tiledb.Array, indexer_cls: Type[TraceIndexer]):
        super().__init__(tdb, indexer_cls(tdb.shape))

    @singledispatchmethod
    def __getitem__(self, i: object) -> None:
        raise TypeError(f"Cannot index by {i.__class__}")

    @__getitem__.register(int)
    @__getitem__.register(np.integer)
    def _get_one(self, i: Int) -> Field:
        return cast(Field, self[ensure_slice(i)][0])

    @__getitem__.register(slice)
    def _get_many(self, i: slice) -> List[Field]:
        bounding_box, post_reshape_indices = self._indexer[i]
        header_arrays = self._tdb[bounding_box]
        keys = header_arrays.keys()
        columns = [header_arrays[key].reshape(-1)[post_reshape_indices] for key in keys]
        return [dict(zip(keys, row)) for row in zip(*columns)]


class Attributes(TraceIndexable):
    def __init__(self, tdb: tiledb.Array, indexer_cls: Type[TraceIndexer]):
        super().__init__(tdb, indexer_cls(tdb.shape))

    def __getitem__(self, i: Index) -> np.ndarray:
        bounding_box, post_reshape_indices = self._indexer[ensure_slice(i)]
        return self._tdb[bounding_box].reshape(-1)[post_reshape_indices]


class Line:
    def __init__(
        self,
        dim_name: str,
        labels: np.ndarray,
        offsets: np.ndarray,
        tdb: tiledb.Array,
    ):
        self.name = dim_name
        self._tdb = tdb
        self._label_indexer = LabelIndexer(labels)
        self._offset_indexer = LabelIndexer(offsets)
        self._default_offset = offsets[0]

    def __len__(self) -> int:
        return cast(int, self._tdb.shape[self._dims.index(self.name)])

    def __getitem__(self, i: Union[Index, Tuple[Index, Index]]) -> np.ndarray:
        if isinstance(i, tuple):
            labels, offsets = i
        else:
            labels, offsets = i, self._default_offset
        data = self._tdb[self._get_tdb_indices(labels, offsets)]
        data = self._moveaxis(data, labels, offsets)
        return data

    _dims = property(lambda self: [dim.name for dim in self._tdb.schema.domain])

    def _get_tdb_indices(self, labels: Index, offsets: Index) -> Tuple[Index, ...]:
        dims = self._dims
        composite_index = [slice(None)] * self._tdb.ndim
        composite_index[dims.index(self.name)] = self._label_indexer[labels]
        composite_index[dims.index("offsets")] = self._offset_indexer[offsets]
        return tuple(composite_index)

    def _moveaxis(self, data: np.ndarray, labels: Index, offsets: Index) -> np.ndarray:
        labels_dim = self.name
        offsets_dim = "offsets"

        dims = self._dims
        if not isinstance(labels, slice):
            dims.remove(labels_dim)
        if not isinstance(offsets, slice):
            dims.remove(offsets_dim)

        # ensure the labels dim is the first axis (if present)
        try:
            labels_axis = dims.index(labels_dim)
            if labels_axis != 0:
                data = np.moveaxis(data, labels_axis, 0)
                labels_axis = 0
        except ValueError:
            labels_axis = -1

        # ensure the offsets dim is right after the labels
        try:
            offsets_axis = dims.index(offsets_dim)
            if offsets_axis != labels_axis + 1:
                data = np.moveaxis(data, offsets_axis, labels_axis + 1)
        except ValueError:
            pass

        return data


NestedFieldList = Union[List[Field], List[List[Field]], List[List[List[Field]]]]


class HeaderLine(Line):
    def __getitem__(self, i: Union[Index, Tuple[Index, Index]]) -> NestedFieldList:
        if isinstance(i, tuple):
            labels, offsets = i
        else:
            labels, offsets = i, self._default_offset
        header_dicts = self._tdb[self._get_tdb_indices(labels, offsets)]
        header_keys = header_dicts.keys()
        data = np.stack(header_dicts.values())
        data = np.moveaxis(data, 0, -1)
        data = self._moveaxis(data, labels, offsets)
        data = np.apply_along_axis(lambda v: dict(zip(header_keys, v)), -1, data)
        return cast(NestedFieldList, data.tolist())


class Depth:
    def __init__(self, tdb: tiledb.Array):
        self._tdb = tdb

    def __len__(self) -> int:
        return cast(int, self._tdb.shape[-1])

    def __getitem__(self, i: Index) -> np.ndarray:
        ndim = self._tdb.ndim
        if ndim > 2:
            # segyio doesn't currently support offset indexing for depth; always selects the first
            # https://github.com/equinor/segyio/issues/474
            data = self._tdb[..., 0, i]
            ndim -= 1
        else:
            data = self._tdb[..., i]
        # move samples as first dimension
        return np.moveaxis(data, -1, 0) if data.ndim == ndim else data


class Gather:
    def __init__(
        self,
        ilines: np.ndarray,
        xlines: np.ndarray,
        offsets: np.ndarray,
        tdb: tiledb.Array,
    ):
        self._tdb = tdb
        self._iline_indexer = LabelIndexer(ilines)
        self._xline_indexer = LabelIndexer(xlines)
        self._offset_indexer = LabelIndexer(offsets)
        self._default_offset = offsets[0] if len(offsets) == 1 else slice(None)

    def __getitem__(self, t: Tuple[Index, ...]) -> np.ndarray:
        if len(t) == 3:
            ilines, xlines, offsets = t
        else:
            ilines, xlines = t
            offsets = self._default_offset

        dims = tuple(dim.name for dim in self._tdb.schema.domain)
        composite_index = [None] * 3
        composite_index[dims.index("ilines")] = self._iline_indexer[ilines]
        composite_index[dims.index("xlines")] = self._xline_indexer[xlines]
        composite_index[dims.index("offsets")] = self._offset_indexer[offsets]

        data = self._tdb[tuple(composite_index)]
        # segyio returns always (ilines, xlines); convert from (fast, slow) if necessary
        if (
            dims[0] == "xlines"
            and isinstance(ilines, slice)
            and isinstance(xlines, slice)
        ):
            data = data.swapaxes(0, 1)
        return data



import itertools as it
from functools import singledispatch
from typing import TYPE_CHECKING, List, Tuple, Union

import numpy as np

from .singledispatchmethod import singledispatchmethod  # type: ignore

# https://github.com/python/typing/issues/684#issuecomment-548203158
if TYPE_CHECKING:  # pragma: nocover
    from enum import Enum

    class ellipsis(Enum):
        Ellipsis = "..."

    Ellipsis = ellipsis.Ellipsis
else:
    ellipsis = type(Ellipsis)


Int = Union[int, np.integer]
Index = Union[Int, slice]


class TraceIndexer:
    def __init__(self, shape: Tuple[int, ...]):
        self._shape = shape

    def __len__(self) -> int:
        return int(np.asarray(self._shape).prod())

    def __getitem__(
        self, trace_index: Index
    ) -> Tuple[Tuple[Index, ...], Union[int, List[int], ellipsis]]:
        """
        Given a trace index, return a `(bounding_box, post_reshape_indices)` tuple where:
        - `bounding_box` is a tuple of (int or slice) indices for each dimension in shape that
          enclose all data of the requested `trace_index`.
        - `post_reshape_indices` is a list of indices to select from the reshaped 1-dimensional
          bounding box in order to get the requested `trace_index` data. It may also be ellipsis
          (...) if the whole bounding box is to be selected.
        """
        return (trace_index,), Ellipsis


class StructuredTraceIndexer(TraceIndexer):
    @singledispatchmethod
    def __getitem__(self, i: object) -> None:
        raise TypeError(f"Cannot index by {i.__class__}")

    @__getitem__.register(int)
    @__getitem__.register(np.integer)
    def _get_one(self, trace_index: Int) -> Tuple[Tuple[int, ...], ellipsis]:
        return np.unravel_index(trace_index, self._shape), Ellipsis

    @__getitem__.register(slice)
    def _get_many(self, trace_index: slice) -> Tuple[Tuple[slice, ...], List[int]]:
        # get indices in 1D (trace index) and 3D (fast-slow-offset indices)
        raveled_indices = np.arange(len(self))[trace_index]
        unraveled_indices = np.unravel_index(raveled_indices, self._shape)
        unique_unraveled_indices = tuple(map(np.unique, unraveled_indices))
        try:
            bounding_box = tuple(map(ensure_slice, unique_unraveled_indices))
        except ValueError:
            raise NotImplementedError(
                "Multi-range subarrays are not currently supported"
            )

        # find the requested subset of indices from the cartesian product
        points = frozenset(zip(*unraveled_indices))
        post_reshape_indices = [
            i
            for i, point in enumerate(it.product(*unique_unraveled_indices))
            if point in points
        ]
        return bounding_box, post_reshape_indices


class LabelIndexer:
    def __init__(self, labels: np.ndarray):
        if not issubclass(labels.dtype.type, np.integer):
            raise ValueError("labels should be integers")
        if len(np.unique(labels)) != len(labels):
            raise ValueError(f"labels should not contain duplicates: {labels}")
        self._labels = labels
        self._min_label = int(labels.min())
        self._max_label = int(labels.max() + 1)
        self._sorter = labels.argsort()

    @singledispatchmethod
    def __getitem__(self, i: object) -> None:
        raise TypeError(f"Cannot index by {i.__class__}")

    @__getitem__.register(int)
    @__getitem__.register(np.integer)
    def _get_one(self, label: Int) -> int:
        indices = np.flatnonzero(label == self._labels)
        assert indices.size <= 1, indices
        if indices.size == 0:
            raise ValueError(f"{label} is not in labels")
        return int(indices[0])

    @__getitem__.register(slice)
    def _get_many(self, label_slice: slice) -> slice:
        return ensure_slice(self._label_slice_to_indices(label_slice))

    def _label_slice_to_indices(self, label_slice: slice) -> np.ndarray:
        start, stop, step = label_slice.start, label_slice.stop, label_slice.step
        min_label = self._min_label
        if step is None or step > 0:  # increasing step
            if start is None or start < min_label:
                start = min_label
        else:  # decreasing step
            if stop is None or stop < min_label - 1:
                stop = min_label - 1

        label_range = np.arange(*slice(start, stop, step).indices(self._max_label))
        indices = self._sorter[
            self._labels.searchsorted(label_range, sorter=self._sorter)
        ]
        return indices[self._labels[indices] == label_range]


@singledispatch
def ensure_slice(obj: object) -> slice:
    raise TypeError(f"Cannot convert {obj.__class__} to slice")


@ensure_slice.register(slice)
def ensure_slice_identity(s: slice) -> slice:
    return s


@ensure_slice.register(int)
@ensure_slice.register(np.integer)
def ensure_slice_int(i: Int) -> slice:
    return slice(i, i + 1)


@ensure_slice.register(np.ndarray)
def ensure_slice_array(a: np.ndarray) -> slice:
    if not issubclass(a.dtype.type, np.integer):
        raise ValueError("Non-integer array cannot be converted to slice")
    if a.ndim > 1:
        raise ValueError(f"{a.ndim}D array cannot be converted to slice")
    if a.ndim == 1 and len(a) == 0:
        raise ValueError("Empty array cannot be converted to slice")
    if a.ndim == 0 or len(a) == 1:
        return ensure_slice(a.item())

    diffs = a[1:] - a[:-1]
    if not (np.all(diffs > 0) or np.all(diffs < 0)):
        raise ValueError(
            "Non-monotonically increasing or decreasing array cannot be converted to slice"
        )
    unique_diffs = np.unique(diffs)
    if len(unique_diffs) > 1:
        raise ValueError("Non-range array cannot be converted to a slice")

    start = a[0]
    step = unique_diffs[0]
    stop = a[-1] + (1 if step > 0 else -1)
    return slice(start, stop, step)



from pathlib import Path

from setuptools import setup

with Path(__file__).parent.joinpath("README.md").open() as f:
    long_description = f.read()

setup(
    name="tilesegy",
    version="0.1.0",
    description="Python library for fast access to seismic data using TileDB",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gsakkis/tilesegy",
    author="George Sakkis",
    author_email="george.sakkis@gmail.com",
    packages=["tilesegy"],
    install_requires=["cached_property", "segyio>=1.9.3", "tiledb"],
    setup_requires=["pytest-runner"],
    tests_require=["pytest", "pytest-coverage", "pytest-xdist", "filelock"],
    entry_points={"console_scripts": ["segy2tiledb=tilesegy.cli:main"]},
    license="MIT",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
    ],
)



# type: ignore

"""Backport of @functools.singledispatchmethod to Python 3.4-3.7."""

import functools

try:
    singledispatchmethod = functools.singledispatchmethod
except AttributeError:

    class singledispatchmethod:
        """Single-dispatch generic method descriptor."""

        def __init__(self, func):
            if not callable(func) and not hasattr(func, "__get__"):
                raise TypeError("{!r} is not callable or a descriptor".format(func))

            self.dispatcher = functools.singledispatch(func)
            self.func = func

        def register(self, cls, method=None):
            return self.dispatcher.register(cls, func=method)

        def __get__(self, obj, cls):
            def _method(*args, **kwargs):
                method = self.dispatcher.dispatch(args[0].__class__)
                return method.__get__(obj, cls)(*args, **kwargs)

            _method.__isabstractmethod__ = self.__isabstractmethod__
            _method.register = self.register
            functools.update_wrapper(_method, self.func)
            return _method

        @property
        def __isabstractmethod__(self):
            return getattr(self.func, "__isabstractmethod__", False)
